/*  Send85 - Command-line utility to send data to TI85
    Written by Magnus Svedin.
    Thanks to Per Finander, Magnus Hagander, Gary Marsh, David Boozer,
    Jani Halme and all the people who's made the djgpp c-compiler. 

Known bugs:
    * Makes an undefined error when pressing skip/exit when mem full
    * Unable to open file without reason
*/

/* #define Linux */
#define Dos

#ifdef Linux
#define SwitchChar '-'
#define getch() getchar()
#endif
#ifdef Dos
#define SwitchChar '/'
#endif

#define uchar unsigned char
#define uint unsigned int
#define bool uchar
#define true 1
#define false 0

#include <stdio.h>
#ifdef Dos
#include <conio.h>
#endif
uchar *tmp, *tmp2;    /* Data needed in c85.h BEFORE that include */
uchar *data;
uint PORT=0x378;
uint PORTTYPE=0x0000;   /* 0=com-port, other=lpt-port , must be uint because of defport.com */
int total_data=-1;
FILE *f;
bool showstatus=true;
bool do_wait=true;
#ifdef Linux
#include "c85linux.h"
#endif
#include "c85.h"

void SendVarFile(char *path, char *FileName);
void SendBackupFile(char *path, char *FileName);

int main(int numarg, char *arg[])
{
  int a;
#ifdef Dos
  struct ffblk curblk;
  char path[255];
  setvbuf(stdout, NULL, _IONBF, NULL); /* Turn off screen-buffering */
#endif
  tmp=(uchar *)malloc(10);    /* used in all functions to save the bother of */
  tmp2=(uchar *)malloc(10);   /* allocation and deallocation                 */
  data=(uchar *)malloc(32000);
  if(numarg==1)    /* if no argument specified display helptext */
  {
    printf("Send85 v2.1   Utility to send files to TI85. By Magnus Svedin\n");
    printf("Usage: send85 [file(s)] [%cC[com-adr]] [%cL[lpt-adr] [%cQ] [%cS]\n", SwitchChar, SwitchChar, SwitchChar, SwitchChar);
    printf("       %cC     Use com-port with specified addr (value in hex)\n", SwitchChar);
    printf("       %cL     Use lpt-port with specified addr (value in hex)\n", SwitchChar);
    printf("       %cQ     Quiet/Quick mode, doesn't display send-status\n", SwitchChar);
    printf("       %cS     Skip pause before start\n", SwitchChar);
    if(PORTTYPE!=0)
      printf("       Default port is LPT at adr %xh.\n", PORT);
    else
      printf("       Default port is COM at adr %xh.\n", PORT);
    exit(1);
  }
  printf("Send85 v2.1 by Magnus Svedin. Press ESC to cancel transfer.\n");
  for(a=1;a<numarg;a++)     /* Checking for switches */
  {
    if(arg[a][0]==SwitchChar)
    {
      /*strlwr(arg[a]);*/
      switch(arg[a][1])
      {
        case 'c':
	  sscanf((char *)(arg[a]+2), "%x", &PORT);
	  PORTTYPE=0;
	  break;
	case 'C':
	  sscanf((char *)(arg[a]+2), "%x", &PORT);
	  PORTTYPE=0;
	  break;
	case 'l':
	  sscanf((char *)(arg[a]+2), "%x", &PORT);
	  PORTTYPE=1;
	  break;
	case 'L':
	  sscanf((char *)(arg[a]+2), "%x", &PORT);
	  PORTTYPE=1;
	  break;
	case 'q':showstatus=false;break;
	case 's':do_wait=false;break;
	case 'Q':showstatus=false;break;
	case 'S':do_wait=false;break;
	default:
    printf("Ignoring illegal switch: %s\n", arg[a]);
      }
    }
  }
  initport();
  if(do_wait)
  {
#ifdef Linux
    printf("Set calculator in receive-mode and press enter.\n");
#else
    printf("Set calculator in receive-mode and press any key.\n");
#endif
    getch();
  }
#ifdef Dos
  for(a=1;a<numarg;a++)
  {
    if(arg[a][0]!=SwitchChar)
    {
      tmp[1]=strlen(arg[a])+1;
      for(tmp[0]=0;tmp[0]<strlen(arg[a]);tmp[0]++)
      if(arg[a][tmp[0]]=='\\')
        tmp[1]=tmp[0];
      if(tmp[1]!=strlen(arg[a])+1)
	strncpy(path, arg[a], tmp[1]+1);
      path[tmp[1]+1]='\0';
      if(findfirst(arg[a], &curblk, 0)==0)
      {
        if(curblk.ff_name[strlen(curblk.ff_name)-1]=='B')
	  SendBackupFile(path, curblk.ff_name);
        else
	  SendVarFile(path, curblk.ff_name);
	while(findnext(&curblk)==0)
	  if(curblk.ff_name[strlen(curblk.ff_name)-1]=='B')
	    SendBackupFile(path, curblk.ff_name);
	  else
	    SendVarFile(path, curblk.ff_name);
      }
      else
  printf("%s: No such file.\n", arg[a]);
    }
  }
#endif
#ifdef Linux
  for(a=1;a<numarg;a++)
  {
    if(arg[a][0]!=SwitchChar)
    {
      if(arg[a][strlen(arg[a])-1]=='b' || arg[a][strlen(arg[a])-1]=='B')
        SendBackupFile("", arg[a]);
      else
	SendVarFile("", arg[a]);
    }
  }
#endif
  if(total_data!=-1)
  {
    SendByte(0x05);
    SendByte(0x92);
    SendByte(total_data & 0xff);
    SendByte((total_data & 0xff00) >> 8);
    GetByte();
    GetByte();
    GetByte();
    GetByte();
  }
  free(data);
  free(tmp);
  free(tmp2);
  return(0);
}

void SendVarFile(char *path, char *FileName)
{
  int a, FileLen;
  char VarName[16];
  uchar dosend;
  uint chsum, dl, skipped;
  printf("%c%s: ", 13, FileName);
  strcpy(data, path);
  strcat(data, FileName);               
  if((f=fopen(data, "rb"))==NULL)
  {
    printf("Unable to open file\n");
    return;     
  }
  fread(tmp, 8, 1, f);
  if(strncmp((char*)tmp, "**TI85**", 8)!=0)
  {
    fclose(f);
    printf("Not a valid .85-file\n");
    return;
  }
  fseek(f, 0, SEEK_END);
  FileLen=(int)ftell(f);
  fseek(f, 0x39, SEEK_SET);
  while(ftell(f)+0x0a < FileLen)
  {
    skipped=0;
    printf("%c%s: Initializing transfer...                          ", 13, FileName);
    for(a=0;a<3000;a++)     /* A pause to make djgpp print the text */
    printf(" %c", 8);
    fread(tmp, 1, 4, f);
    fread(data, 1, tmp[3], f);
    for(a=0, chsum=0;a<4;a++)
    chsum+=(uchar)tmp[a];
    for(a=0;a<tmp[3];a++)
    {
      VarName[a]=data[a];
      chsum+=(uchar)data[a];
    }
    VarName[a]='\0';
    dosend=1;
    while(dosend)   /* Sending var-header */
    {
      if(total_data==-1)
      total_data=0;
      SendByte(0x05);
      SendByte(0x06);
      SendByte(4+(uchar)tmp[3]);
      SendByte(0);
      for(a=0;a<4;a++)
        SendByte((uchar)tmp[a]);
      for(a=0;a<tmp[3];a++)
        SendByte((uchar)data[a]);
      SendByte((uchar)(chsum&0xff));
      SendByte((uchar)((chsum&0xff00)>>8));
      for(a=0;a<4;a++)
        tmp2[a]=GetByte();
      if(tmp2[1]!=0x5a)
       dosend=false;
    }
    for(a=0;a<4;a++)
    tmp[a]=GetByte();
    if(tmp[1]!=0x09)
    {
      if(tmp[1]==0x36 && tmp[2]==0x01)
      {
	tmp[4]=GetByte();
	if(tmp[4]==0x03)
	{
    printf("\nOut of memory in TI85.\n");
          fclose(f);
          exit(1);
	}
	if(tmp[4]==0x01)
	{
    printf("\nBreak from TI85 signaled.\n");
	  GetByte();
	  GetByte();
          fclose(f);
          exit(1);
	}
	if(tmp[4]==0x02)
	{
	  GetByte();
	  GetByte();
          printf("%c%s: %s skipped.                               \n", 13, FileName, VarName);
          skipped=1;
          fread(tmp, 1, 2, f);
          dl=(uint)(uchar)tmp[0];
          dl+=((uint)(tmp[1])) << 8;
          fseek(f, dl+2, SEEK_CUR);
          SendByte(0x05);
          SendByte(0x56);
          SendByte(0x00);
          SendByte(0x00);
        }
      }
      else
      {
        printf("\nUndefined error. (%xh, %xh, %xh, %xh)\n", tmp[0], tmp[1], tmp[2], tmp[3]);
        fclose(f);
        exit(1);
      }
    }
    if(skipped==0)
    {
      fread(tmp, 1, 2, f);       /* Sending data part */
      dl=(uint)(uchar)tmp[0];
      dl+=((uint)(tmp[1])) << 8;
      if(showstatus)
        printf("%c%s: Sending %s (%d bytes). (00%%)                    %c%c%c%c%c%c%c%c%c%c%c%c%c%c%c%c%c%c%c%c", 13, FileName, VarName, dl,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8);
      else
        printf("%c%s: Sending %s (%d bytes).                     ", 13, FileName, VarName, dl);
      total_data+=dl;
      dosend=true;
      while(dosend)
      {
        SendByte(0x05);
        SendByte(0x56);
        SendByte(0x00);
        SendByte(0x00);
        SendByte(0x05);
        SendByte(0x15);
        SendByte(tmp[0]);
        SendByte(tmp[1]);
        fread(data, 1, dl, f);
        chsum=0;
        if(showstatus)
          for(a=0;a<dl;a++)
          {
            chsum+=data[a];
            SendByte(data[a]);
            printf("%c%c%c%c%.2d%%)", 8, 8, 8, 8, a*100/dl);
          }
        else
          for(a=0;a<dl;a++)
          {
            chsum+=data[a];
            SendByte(data[a]);
          }
        SendByte(chsum & 0xff);
        SendByte((chsum & 0xff00) >> 8);
        for(a=0;a<4;a++)
          tmp2[a]=GetByte();
        if(tmp2[1]==0x56)
          dosend=false;
      }
      printf("%c%s: %s (%d bytes) Done.                                 \n", 13, FileName, VarName, dl);
      fread(tmp, 1, 2, f);    /* to position filepointer in correct position */
    }
  }
  fclose(f);
}

void SendBackupFile(char *path, char *FileName)
{
	int a, part;
	ulong chsum;
	uint dl, filebytes, sentbytes;
	uchar dosend;
  printf("%c%s: ", 13, FileName);
	strcpy(data, path);
	strcat(data, FileName);
	/* fcloseall(); */
	if((f=fopen(data, "rb"))==NULL)
	{
    printf("Unable to open file\n");
		return;     
	}
	fread(tmp, 8, 1, f);
	if(strncmp((char*)tmp, "**TI85**", 8)!=0)
	{
		fclose(f);
    printf("Not a valid .85-file\n");
		return;
	}

	fseek(f, 0x35, SEEK_SET);
	fread(tmp, 2, 1, f);
	filebytes=(uint)(uchar)tmp[0];
	filebytes+=((uint)(tmp[1])) << 8;
	fread(tmp, 2, 1, f);
	dl=(uint)(uchar)tmp[0];
	dl+=((uint)(tmp[1])) << 8;
	fread(data, 1, dl, f);
	dosend=1;
  printf("%c%s: Initializing transfer...                          ", 13, FileName);
	while(dosend)    /* sending part 1  */
	{
		SendByte(0x05);
		SendByte(0x06);
		SendByte(tmp[0]);
		SendByte(tmp[1]);
		chsum=0;
		for(a=0;a<dl;a++)
		{
			chsum=chsum+data[a];
			SendByte(data[a]);
		}
		SendByte((ulong)(chsum&0xff));
		SendByte(((ulong)(chsum&0xff00))>>8);
		for(a=0;a<4;a++)    
			tmp2[a]=GetByte();
		if(tmp2[1]==0x56)
			dosend=0;
	}
	sentbytes=dl;
	for(a=0;a<3000;a++);    /* TI85 doesn't seem to keep up with the cpu here */
	for(a=0;a<4;a++)
		tmp[a]=GetByte();
	if(tmp[1]!=0x09)
	{
    printf("\nBreak from TI85 signaled.\n");
		return;
	}
	SendByte(0x05);
	SendByte(0x56);
	SendByte(tmp[2]);
	SendByte(tmp[3]);
	if(showstatus)
    printf("%c%s: Sending %d bytes. (00%%)                    %c%c%c%c%c%c%c%c%c%c%c%c%c%c%c%c%c%c%c%c", 13, FileName, filebytes,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8);
	else
    printf("%c%s: Sending %d bytes.                     ", 13, FileName, filebytes);
	for(a=0;a<3000;a++)     /* A do_wait to make djgpp print the text */
    printf(" %c", 8);
	for(part=0;part<3;part++)
	{
		fread(tmp, 1, 2, f);     /* reading parts */
		dl=(uint)(uchar)tmp[0];
		dl+=(((uint)((uchar)tmp[1])) << 8);
		fread(data, 1, dl, f);
		dosend=1;
		while(dosend)                           /* sending parts */
		{
			SendByte(0x05);
			SendByte(0x15);
			SendByte(tmp[0]);
			SendByte(tmp[1]);
			if(showstatus)
	for(a=0, chsum=0;a<dl;a++)
	{
		chsum+=data[a];
		SendByte(data[a]);
    printf("%c%c%c%c%.2d%%)", 8, 8, 8, 8, (a+sentbytes)*100/filebytes);
	}
			else
	for(a=0, chsum=0;a<dl;a++)
	{
		chsum+=data[a];
		SendByte(data[a]);
	}
			SendByte((ulong)(chsum&0xff));
			SendByte(((ulong)(chsum&0xff00))>>8);
			for(a=0;a<4;a++)
				tmp2[a]=GetByte();
			if(tmp2[1]==0x56)
				dosend=0;
		}
		sentbytes+=dl;
	}
  printf("%c%s: %d bytes Done.                                             \n", 13, FileName, filebytes);
	fclose(f);
	total_data=-1;
	return;
}
